// Operation to find a clustering of a given graph with high modularity.
//
// This implementation is based on iteratively doing small tweaks on the clustering based on the view
// visible in the individual partitions of data and then reshuffle.
//
// This operation ignores the direction of edges.

package com.lynxanalytics.biggraph.graph_operations

import org.apache.spark
import org.apache.spark.rdd.RDD
import scala.collection.mutable
import scala.util.Random

import com.lynxanalytics.biggraph.{bigGraphLogger => log}
import com.lynxanalytics.biggraph.graph_api._
import com.lynxanalytics.biggraph.spark_util.Implicits._

object FindModularClusteringByTweaks extends OpFromJson {
  private val maxIterationsParameter = NewParameter("maxIterations", -1)
  private val minIncrementPerIterationParameter = NewParameter("minIncrementPerIteration", 0.001)
  class Input extends MagicInputSignature {
    val (vs, edges) = graph
    val weights = edgeAttribute[Double](edges)
  }
  class Output(implicit instance: MetaGraphOperationInstance, inputs: Input) extends MagicOutput(instance) {
    val clusters = vertexSet
    val belongsTo = edgeBundle(
      inputs.vs.entity,
      clusters,
      properties = EdgeBundleProperties.partialFunction)
  }
  def fromJson(j: JsValue) =
    FindModularClusteringByTweaks(
      maxIterationsParameter.fromJson(j),
      minIncrementPerIterationParameter.fromJson(j))

  case class ClusterData(
      // The sum of degrees of all nodes in this cluster. Note, if an edge goes within the cluster
      // then its weight is counted twice in this sum - once for both its endpoints.
      degreeSum: Double,
      // The sum of degrees of all nodes in this cluster in the subgraph induced by the nodes of
      // this cluster. In other words, this is two times the sum of weights of edges that go
      // within this cluster.
      insideDegreeSum: Double,
      size: Int) {

    def add(connection: Double, other: ClusterData): ClusterData = {
      ClusterData(
        degreeSum + other.degreeSum,
        insideDegreeSum + other.insideDegreeSum + 2 * connection,
        size + other.size)
    }
    def remove(connection: Double, other: ClusterData): ClusterData = {
      ClusterData(
        degreeSum - other.degreeSum,
        insideDegreeSum - other.insideDegreeSum - 2 * connection,
        size - other.size)
    }
    def addNode(degree: Double, inClusterEdges: Double): ClusterData = {
      ClusterData(
        degreeSum + degree,
        // We add inClusterEdges only once here as it will be also added when processing the
        // other end of the edge.
        insideDegreeSum + inClusterEdges,
        size + 1,
      )
    }

    def modularity(totalDegreeSum: Double) =
      insideDegreeSum / totalDegreeSum -
        degreeSum * degreeSum / totalDegreeSum / totalDegreeSum
  }
  object ClusterData {
    def fromMembers(members: Set[ID], edgeLists: Map[ID, Iterable[(ID, Double)]]): ClusterData = {
      var degreeSum = 0.0
      var insideDegreeSum = 0.0
      for (member <- members) {
        for ((other, weight) <- edgeLists(member)) {
          degreeSum += weight
          if (members.contains(other)) {
            insideDegreeSum += weight
          }
        }
      }
      ClusterData(degreeSum, insideDegreeSum, members.size)
    }
  }

  /* Spectral representation for a cluster within a larger graph.
   *
   * In essence, this class represents an indexing of the members of the cluster
   * v_0, v_1, ..., v_{size-1} and a matrix M. M has the following property. Assume we partition
   * the cluster in two partitions, A and B. Prepare a vector s, such that s_i = 1 if v_i \in A and
   * s_i = -1 if v_i \in B. Then the modularity change if we split the cluster into the two clusters
   * A-B is:
   * s^T M s
   *
   * For a normalized eigenvector w with eigenvalue \lambda we have w^T M w = \lambda. Any vector v
   * can be expressed as a linear combination of orthonormal eigenvectors, v = \sum a_i w_i. Then:
   * v^T M v = \sum a_i^2 \lambda_i
   * This shows that to achieve the largest value for v^T M v, we have to choose an eigenvector
   * with the largest possible eigenvalue.
   * Unfortunately, that normally won't be a -1/+1 vector. So to actually get a split candidate,
   * we take this maximally positive eigenvector, and replace <0 numbers with -1 and >=0 numbers
   * with 1.
   *
   * For details, see: http://www.pnas.org/content/103/23/8577.long
   */
  class ClusterSpectrum(
      totalDegreeSum: Double,
      data: ClusterData,
      members: Set[ID],
      degrees: Map[ID, Double],
      edgeLists: Map[ID, Iterable[(ID, Double)]]) {

    // Keep only edges within the cluster.
    val restrictedEdgeLists = edgeLists
      .filterKeys(id => members.contains(id))
      .mapValues(edges => edges.filter { case (otherId, weight) => members.contains(otherId) })

    // Indexing and reverse indexing.
    val indexedIds = members.toArray
    val idToIndex = indexedIds.zipWithIndex.toMap

    // Matrix M as explained above will be represented as
    // degreeArray * degreeArray^T + sparseMatrix
    // for more efficient matrix multiplication.
    val degreeArray = indexedIds.map(id => degrees(id))

    // See the above article for details, but sparseMatrix is basically the weighted adjacency
    // matrix of the graph restricted to the cluster. There is an additional correction
    // applied to the diagonal elements that makes sure all rows (and columns) sum up to 0.
    val sparseMatrix: Array[Array[(Int, Double)]] = indexedIds.map { id =>
      val insideConnections = restrictedEdgeLists
        .getOrElse(id, List())
        .map { case (id, weight) => (idToIndex(id), weight) }
      val correction =
        data.degreeSum * degrees(id) / totalDegreeSum - insideConnections.map(_._2).sum
      Array((idToIndex(id), correction)) ++ insideConnections
    }

    def dot(v1: Array[Double], v2: Array[Double]): Double = {
      v1.zip(v2).map { case (a, b) => a * b }.sum
    }

    // Computes Mv.
    def multiply(v: Array[Double]): Array[Double] = {
      val degreeDot = dot(degreeArray, v)
      val degreeContribution =
        degreeArray.map(-_ * degreeDot / totalDegreeSum)
      val matrixContribution =
        sparseMatrix.map(line => line.map { case (idx, weight) => v(idx) * weight }.sum)
      degreeContribution.zip(matrixContribution).map { case (a, b) => a + b }
    }

    def normalize(v: Array[Double]): Unit = {
      val norm = Math.sqrt(v.iterator.map(x => x * x).sum)
      for (i <- v.indices) {
        v(i) /= norm
      }
    }

    // Represents the split (part1, members \ part1) as a +1/-1 vector.
    def toPlusMinus(part1: Set[ID]): Array[Double] = {
      indexedIds.map(id => if (part1.contains(id)) 1.0 else -1.0)
    }

    // Tries to find an eigenvector with maximal eigenvalue. In essence, we use the power
    // iteration method as explained here:
    // http://en.wikipedia.org/wiki/Power_iteration
    // The slight issue with that is that it will find the eigenvalue
    // (and a corresponding eigenvector) with the largest possible _absolute_ value.
    // In our case it's very often going to be negative, as you can screw up modularity
    // a lot with a suitable cut... So if we end up with an eigenvalue dominantEV < 0, then
    // we do another power iteration for the matrix M -(2/3)*dominantEV*I. This matrix has the
    // exact same eigenvectors but all eigenvalues are increased by -(2/3)*dominantEV. Now the
    // eigenvalue with the biggest absolute value is going to be positive, and it will correspond
    // to the largest positive eigenvalue of the original matrix M.
    // This modification to the power iteration is not explained in the original article nor did
    // I find this in any official source, so there may be a better way to do this.
    def findBestEigenVector(rnd: Random): Array[Double] = {
      val start = indexedIds.map(_ => rnd.nextDouble() * 2 - 1)
      normalize(start)
      var current = start
      var currentValue = Double.MinValue
      var prev = current
      var prevValue = currentValue
      var it = 0
      do {
        prev = current
        prevValue = currentValue
        current = multiply(current)
        currentValue = dot(current, prev)
        normalize(current)
        it += 1
      } while ((it < 100) && (Math.abs(currentValue - prevValue) > 0.001))
      val dominantEV = dot(current, multiply(current))
      if (dominantEV < 0) {
        // The dominant eigenvalue is negative. Let's reiterate with the matrix:
        // M - (2/3) * dominantEV * I.
        current = start
        var currentValue = Double.MinValue
        do {
          prev = current
          prevValue = currentValue
          current = multiply(current).zip(current)
            .map { case (cv, pv) => cv - pv * dominantEV * 2.0 / 3.0 }
          normalize(current)
          currentValue = dot(current, multiply(current))
          it += 1
        } while ((it < 200) && (Math.abs(currentValue - prevValue) > 0.001))
      }
      current
    }

    // Returns (part1, part2, modularityDelta) where part1 and part2 is a partitioning of
    // the cluster into two parts and modularityDelta is the increase in modularity if
    // we split the cluster into these two parts.
    def bestSplit(rnd: Random): (Set[ID], Set[ID], Double) = {
      val eigenVector = findBestEigenVector(rnd)
      val asPlusMinus = eigenVector.map(x => if (x < 0) -1.0 else 1.0)
      val modularityDelta = dot(asPlusMinus, multiply(asPlusMinus)) / totalDegreeSum / 2.0
      val withGroupLabels = indexedIds.zip(asPlusMinus)
      (
        withGroupLabels.filter(_._2 > 0).map(_._1).toSet,
        withGroupLabels.filter(_._2 < 0).map(_._1).toSet,
        modularityDelta)
    }
  }

  // Returns the change in modularity if the two given clusters were to be merged.
  // In other words, returns:
  //   (cluster1.add(cluster2).modularity - cluter1.modularity - cluster2.modularity)
  def mergeModularityChange(
      totalDegreeSum: Double,
      cluster1: ClusterData,
      cluster2: ClusterData,
      connection: Double): Double = {

    val totalDegreeSumSquare = totalDegreeSum * totalDegreeSum
    2 * connection / totalDegreeSum -
      2 * cluster1.degreeSum * cluster2.degreeSum / totalDegreeSumSquare
  }

  // Computes merge modularity changes for all clusters connected to a given partion.
  // It returns a map where keys are ids of merge canidate clusters and values are
  // (mergeModularityChange, connection) pairs.
  def getMergeCandidates(
      totalDegreeSum: Double,
      clusters: scala.collection.Map[ID, ClusterData],
      cluster: ClusterData,
      members: scala.collection.Set[ID],
      edgeLists: Map[ID, Iterable[(ID, Double)]],
      containedIn: mutable.Map[ID, ID]): Map[ID, (Double, Double)] = {

    val clusterConnections = mutable.Map[ID, Double]().withDefaultValue(0.0)
    for (vertex <- members) {
      for ((otherId, weight) <- edgeLists(vertex)) {
        if (!members.contains(otherId)) {
          for (otherClusterId <- containedIn.get(otherId)) {
            clusterConnections(otherClusterId) += weight
          }
        }
      }
    }
    clusterConnections
      .map {
        case (id, connection) =>
          (
            id,
            (
              mergeModularityChange(totalDegreeSum, cluster, clusters(id), connection),
              connection))
      }
      .toMap
  }

  def refineClusters(
      totalDegreeSum: Double,
      edgeLists: Map[ID, Iterable[(ID, Double)]],
      containedIn: mutable.Map[ID, ID],
      start: spark.util.DoubleAccumulator,
      end: spark.util.DoubleAccumulator,
      increase: spark.util.DoubleAccumulator,
      rnd: Random): Unit = {

    var localIncrease = 0.0
    val clusters = mutable.Map[ID, ClusterData]()
    // We will only try to split changed clusters.
    val changedClusters = mutable.Set[ID]()
    val degrees = edgeLists.mapValues(edges => edges.map(_._2).sum)
    val loops = edgeLists
      .map { case (id, edges) => id -> edges.find(_._1 == id).map(_._2).getOrElse(0.0) }
    for ((id, edges) <- edgeLists) {
      val clusterId = containedIn(id)
      clusters(clusterId) = clusters
        .getOrElse(clusterId, ClusterData(0, 0, 0))
        .addNode(
          degrees(id),
          edges
            .filter { case (otherId, weight) => containedIn.get(otherId) == Some(clusterId) }
            .map(_._2)
            .sum)
    }
    start.add(clusters.values.map(_.modularity(totalDegreeSum)).sum)

    var changed = false
    var i = 0
    do {
      log.info(s"Doing tweaking nodes subiteration $i")
      i += 1
      changed = false
      for ((id, edges) <- edgeLists) {
        val degree = degrees(id)
        val loop = loops(id)
        val homeClusterId = containedIn(id)
        val homeCluster = clusters(homeClusterId)
        if (homeCluster.size == 1) {
          val candidates = getMergeCandidates(
            totalDegreeSum,
            clusters,
            homeCluster,
            Set(id),
            edgeLists,
            containedIn)
          if (candidates.size > 0) {
            val (bestClusterId, (bestModularityChange, bestClusterConnection)) =
              candidates.maxBy { case (id, (value, connection)) => value }
            if (bestModularityChange > 0) {
              changed = true
              localIncrease += bestModularityChange
              containedIn(id) = bestClusterId
              clusters(bestClusterId) =
                clusters(bestClusterId).add(bestClusterConnection, homeCluster)
              changedClusters.add(bestClusterId)
              clusters.remove(homeClusterId)
              changedClusters.remove(homeClusterId)
            }
          }
        } else {
          val singletonCluster = ClusterData(degree, loop, 1)
          val homeConnection = edges
            .filter(_._1 != id)
            .filter { case (otherId, _) => containedIn.get(otherId) == Some(homeClusterId) }
            .map(_._2)
            .sum
          // We temporarily remove the vertex from his own cluster.
          val homeClusterWithoutMe = homeCluster.remove(homeConnection, singletonCluster)
          clusters(homeClusterId) = homeClusterWithoutMe
          val candidates = getMergeCandidates(
            totalDegreeSum,
            clusters,
            singletonCluster,
            Set(id),
            edgeLists,
            containedIn)
          val finalCandidates =
            if (homeConnection == 0)
              // We don't have ANY connection to our own home. In this case candidates won't
              // contain the score of staying in our own cluster. Let's add that.
              candidates.updated(
                homeClusterId,
                (mergeModularityChange(totalDegreeSum, singletonCluster, homeClusterWithoutMe, 0), 0.0))
            else candidates
          val (bestClusterId, (bestModularityChange, bestClusterConnection)) =
            finalCandidates.maxBy { case (id, (change, connection)) => change }
          val currentValue = finalCandidates(homeClusterId)._1
          if (bestModularityChange > currentValue) {
            // If we are strictly better than in the original cluster we move ...
            changed = true
            localIncrease += bestModularityChange - currentValue
            containedIn(id) = bestClusterId
            clusters(bestClusterId) =
              clusters(bestClusterId).add(bestClusterConnection, singletonCluster)
            changedClusters.add(bestClusterId)
            changedClusters.add(homeClusterId)
            // We've already removed the node from its original cluster, so nothing else to do
            // here.
          } else {
            // otherwise we restore our home cluster.
            clusters(homeClusterId) = homeCluster
          }
        }
      }
    } while (changed)
    i = 0
    val contains = mutable.Map(containedIn.groupBy(_._2).mapValues(_.keySet).toSeq: _*)
    assert(clusters.size == contains.size, s"Size mismatch: ${clusters.size} != ${contains.size}")
    do {
      log.info(s"Doing merging clusters subiteration $i")
      changed = false
      i += 1
      for ((clusterId, cluster) <- clusters) {
        val members = contains(clusterId)
        val candidates = getMergeCandidates(
          totalDegreeSum,
          clusters,
          cluster,
          members,
          edgeLists,
          containedIn)
        if (candidates.size > 0) {
          val (bestClusterId, (bestModularityChange, bestClusterConnection)) =
            candidates.maxBy { case (id, (change, connection)) => change }
          if (bestModularityChange > 0) {
            changed = true
            localIncrease += bestModularityChange
            for (id <- members) {
              containedIn(id) = bestClusterId
            }
            contains(bestClusterId) = contains(bestClusterId) ++ members
            clusters(bestClusterId) =
              clusters(bestClusterId).add(bestClusterConnection, cluster)
            changedClusters.add(bestClusterId)
            clusters.remove(clusterId)
            contains.remove(clusterId)
            changedClusters.remove(clusterId)
          }
        }
      }
    } while (changed)

    // Make sure for all clusters the id of the cluster is the id of one of its members. This
    // can be screwed up in the node tweaking subiteration. It can (accidentally) be fixed in
    // cluster merging, thus we do this check here only.
    // The invariant is required when we start splitting clusters, as on a split we
    // need to come up with a new unique cluster id.
    val clusterIds = changedClusters.toIndexedSeq // Make sure this is not a view.
    for (clusterId <- clusterIds) {
      var members = contains(clusterId)
      while (!members.contains(clusterId) && members.nonEmpty) {
        val newId = members.head
        val data = clusters(clusterId)
        if (contains.contains(newId)) {
          // Our new preferred id is already taken. We switch id with the other guy.
          val otherMembers = contains(newId)
          val otherData = clusters(newId)
          contains(newId) = members
          contains(clusterId) = otherMembers
          clusters(newId) = data
          clusters(clusterId) = otherData
          for (member <- members) {
            containedIn(member) = newId
          }
          for (member <- otherMembers) {
            containedIn(member) = clusterId
          }
          // We continue working on the same cluster id in our while loop, as the cluster who
          // newly has this id might very well be broken now. But we have to update members.
          members = otherMembers
          // Both of these guys should be already changed.
          assert(changedClusters.contains(newId), s"$newId is not in $changedClusters")
          assert(changedClusters.contains(clusterId), s"$clusterId is not in $changedClusters")
        } else {
          clusters(newId) = data
          changedClusters.add(newId)
          clusters.remove(clusterId)
          changedClusters.remove(clusterId)
          contains(newId) = members
          contains.remove(clusterId)
          for (member <- members) {
            containedIn(member) = newId
          }
          members = Set()
        }
      }
    }

    log.info("Splitting clusters")
    val splitQueue = mutable.Queue(changedClusters.toSeq: _*)
    while (splitQueue.nonEmpty) {
      val clusterId = splitQueue.dequeue
      val data = clusters(clusterId)
      val members = contains(clusterId)
      val spectrum = new ClusterSpectrum(
        totalDegreeSum,
        data,
        members.toSet,
        degrees,
        edgeLists)
      val bs = spectrum.bestSplit(rnd)
      val (clust1Prelim, clust2Prelim, modularityDelta) = bs
      if (clust1Prelim.nonEmpty && clust2Prelim.nonEmpty) {
        // Makes sure clusterId is a member of clust1.
        val (clust1, clust2) =
          if (clust1Prelim.contains(clusterId)) (clust1Prelim, clust2Prelim)
          else (clust2Prelim, clust1Prelim)
        assert(clust1.contains(clusterId), s"$clusterId is not in $clust1")
        if (modularityDelta > 0) {
          val clust1Data = ClusterData.fromMembers(clust1, edgeLists)
          val clust2Data = ClusterData.fromMembers(clust2, edgeLists)
          val clust2Id = clust2.head
          clusters(clusterId) = clust1Data
          clusters(clust2Id) = clust2Data
          contains(clusterId) = clust1
          contains(clust2Id) = clust2
          for (member <- clust2) {
            containedIn(member) = clust2Id
          }
          localIncrease += modularityDelta
          splitQueue += clusterId
          splitQueue += clust2Id
        }
      }
    }

    increase.add(localIncrease)
    end.add(clusters.values.map(_.modularity(totalDegreeSum)).sum)
  }

  // We do at least smoothingLength iterations and then we exit if the total modularity increase
  // in the last smoothingLength iterations were small. We don't use only one iteration's score as
  // due to the randomness of the algorithm it's possible to have significant improvements after
  // some non-fruitful iterations. The actual number 5 is a result of some very deep theoretical
  // arguments which this comment is too narrow to contain.
  val smoothingLength = 5
}

import FindModularClusteringByTweaks._
case class FindModularClusteringByTweaks(
    maxIterations: Int,
    minIncrementPerIteration: Double)
    extends SparkOperation[Input, Output] {

  override val isHeavy = true
  @transient override lazy val inputs = new Input
  def outputMeta(instance: MetaGraphOperationInstance) = new Output()(instance, inputs)

  override def toJson =
    maxIterationsParameter.toJson(maxIterations) ++
      minIncrementPerIterationParameter.toJson(minIncrementPerIteration)

  def execute(
      inputDatas: DataSet,
      o: Output,
      output: OutputBuilder,
      rc: RuntimeContext): Unit = {
    implicit val id = inputDatas
    val vs = inputs.vs.rdd
    val vPart = vs.partitioner.get
    val edgeLists = inputs.edges.rdd.sortedJoin(inputs.weights.rdd)
      .map {
        case (id, (e, w)) =>
          if (e.src < e.dst) ((e.src, e.dst), w)
          else ((e.dst, e.src), w)
      }
      .reduceBySortedKey(vPart, _ + _)
      .flatMap {
        case ((v1, v2), w) =>
          if (v1 == v2) Iterator((v1, (v1, 2 * w))) else Iterator((v1, (v2, w)), (v2, (v1, w)))
      }
      .groupBySortedKey(vPart)

    val totalDegreeSum = edgeLists.map { case (id, edges) => edges.map(_._2).sum }.sum

    val numParts = vPart.numPartitions
    // members is a (unique cluster id -> list of vertex ids) mapping.
    // Invariant: one vertex id only shows up once in only one of the vertex id lists.
    var members: RDD[(ID, Iterable[ID])] = edgeLists.mapValuesWithKeys { case (key, _) => Seq(key) }

    var i = 0
    // We keep the last few modularity increment values to decide whether we want to
    // continue with the iterations or not.
    var lastIncrements = List[Double]()
    import FindModularClusteringByTweaks.smoothingLength
    do {
      val start = rc.sparkContext.doubleAccumulator
      val end = rc.sparkContext.doubleAccumulator
      val increase = rc.sparkContext.doubleAccumulator
      val vertexMeta = members
        .mapPartitionsWithIndex({
          case (idx, it) =>
            val seed = new Random((idx << 16) + i).nextLong
            val rnd = new Random(seed)
            it.flatMap {
              case (cid, vs) =>
                val pid = rnd.nextInt(numParts)
                vs.map(vid => vid -> (cid, pid))
            }
        })
        .sortUnique(vPart) // Unique because of the invariant of members, see above.
      val perPartitionData = vertexMeta.sortedJoin(edgeLists)
        .map { case (vid, ((cid, pid), edges)) => pid -> (vid, cid, edges) }
        .sort(vPart)

      val refinedContainedIn = perPartitionData.mapPartitionsWithIndex {
        case (pid, vertexIt) =>
          val seed = new Random((pid << 16) + i).nextLong
          val rnd = new Random(seed)
          val asSeq = vertexIt.toSeq
          val containedIn = mutable.Map(
            asSeq.map { case (_, (vid, cid, edges)) => vid -> cid }: _*)
          val edgeLists = asSeq
            .map { case (_, (vid, cid, edges)) => vid -> edges }
            .toMap
          log.info(s"Starting cluster refinement iteration $i for partition $pid")
          refineClusters(totalDegreeSum, edgeLists, containedIn, start, end, increase, rnd)
          containedIn.iterator
      }
      // in refinedContainedIn:
      // - keys are unique across the whole RDD, but not sorted
      // - vertex ids in the values are unique across the whole RDD
      // TODO: We know all clusters are contained in the same partition, so this could be optimized.
      members = refinedContainedIn.map(_.swap).groupByKey().cache()
      // We explicitly evaluate members to get the value of the accumulators that we need for
      // the termination decision.
      members.foreach(_ => ())
      log.info(
        s"Modularity in iteration $i increased by ${increase.value} " +
          s"from ${start.value} to ${end.value}")
      assert(Math.abs(start.value + increase.value - end.value) < 0.0000001, "Increase mismatch")
      val inc: Double = increase.value
      lastIncrements = (inc +: lastIncrements).take(smoothingLength)
      i += 1
    } while (((lastIncrements.size < smoothingLength) ||
      (lastIncrements.sum > minIncrementPerIteration * smoothingLength)) &&
      ((i < maxIterations) || (maxIterations < 0)))

    val belongsToFromEdges = members
      .flatMap { case (cid, vids) => vids.map(_ -> cid) }
      .sortUnique(vPart)
    val fullMembers = vs.sortedLeftOuterJoin(belongsToFromEdges)
      .map { case (vid, (_, cidOpt)) => cidOpt.getOrElse(vid) -> vid }.groupByKey()
    val clusters = fullMembers.randomNumbered(vPart).mapValues(_._2)
    output(o.clusters, clusters.mapValues(_ => ()))
    output(
      o.belongsTo,
      clusters
        .flatMap { case (p, members) => members.map(m => Edge(m, p)) }
        .randomNumbered(vPart))
  }
}
